use crate::classifier::Classifier;
use crate::data::*;
use crate::userdb::{IsSpam, UserDb};
use std::collections::HashMap;
use std::sync::{Arc, Mutex};

#[derive(Clone)]
pub struct Db {
    userdb: Arc<Mutex<UserDb>>,
    classifier: Arc<Mutex<Classifier>>,
    cache: Arc<Mutex<Cache>>,
}

struct Cache {
    score: HashMap<UserId, f32>,
    tokens: HashMap<UserId, Vec<String>>,
}

impl Db {
    // Creating

    pub fn create(userdb: UserDb, classifier: Classifier) -> Self {
        let cache = Cache::create(&userdb, &classifier);
        Self {
            userdb: Arc::new(Mutex::new(userdb)),
            classifier: Arc::new(Mutex::new(classifier)),
            cache: Arc::new(Mutex::new(cache)),
        }
    }

    pub fn replace_userdb(&self, newdb: UserDb) {
        let userdb: &mut UserDb = &mut self.userdb.lock().unwrap();
        let _ = std::mem::replace(userdb, newdb);
        let new_cache = Cache::create(userdb, &self.classifier.lock().unwrap());
        let cache: &mut Cache = &mut self.cache.lock().unwrap();
        let _ = std::mem::replace(cache, new_cache);
    }

    // Reading

    pub fn with_userdb<F, T>(&self, f: F) -> T
    where
        F: FnOnce(&UserDb) -> T,
    {
        let lock = &self.userdb.lock().unwrap();
        f(lock)
    }

    pub fn login(&self, uid: UserId) -> Option<String> {
        self.with_userdb(|u| u.userdata(uid).map(|d| d.login.clone()))
    }

    pub fn score(&self, uid: UserId) -> Option<f32> {
        self.cache.lock().unwrap().score.get(&uid).copied()
    }

    pub fn with_tokens<F>(&self, uid: UserId, f: F)
    where
        F: FnOnce(Option<&[String]>),
    {
        let lock = self.cache.lock().unwrap();
        f(lock.tokens.get(&uid).map(|v| &**v))
    }

    // Updating

    // pub fn recompute_scores(&self, classifier: &Classifier) {
    //     let lock = &mut self.inner.lock().unwrap();
    //     lock.recompute_scores(classifier)
    // }

    pub fn set_spam(&self, uid: UserId, is_spam: Option<IsSpam>) {
        let udb = &mut self.userdb.lock().unwrap();
        udb.set_spam(uid, is_spam)
    }

    pub fn with_classifier<F, T>(&self, f: F) -> T
    where
        F: FnOnce(&Classifier) -> T,
    {
        let classifier = &self.classifier.lock().unwrap();
        f(classifier)
    }

    pub fn with_classifier_mut<F, T>(&self, f: F) -> T
    where
        F: FnOnce(&mut Classifier) -> T,
    {
        let classifier = &mut self.classifier.lock().unwrap();
        let res = f(classifier);
        // recompute scores
        let cache: &mut Cache = &mut self.cache.lock().unwrap();
        for (id, tokens) in &cache.tokens {
            cache.score.insert(*id, classifier.score(tokens));
        }
        res
    }

    pub fn remove_user(&self, uid: UserId) {
        let userdb = &mut self.userdb.lock().unwrap();
        userdb.remove_user(uid);
        let cache = &mut self.cache.lock().unwrap();
        cache.remove_user(uid);
    }
}

impl Cache {
    fn create(userdb: &UserDb, classifier: &Classifier) -> Self {
        let mut tokens = HashMap::new();
        let mut score = HashMap::new();

        for (id, user, _) in userdb {
            let user_tokens = user.to_tokens();
            let user_score = classifier.score(&user_tokens);
            tokens.insert(id, user_tokens);
            score.insert(id, user_score);
        }

        Cache { tokens, score }
    }

    fn remove_user(&mut self, uid: UserId) {
        self.score.remove(&uid);
        self.tokens.remove(&uid);
    }
}
