use crate::util::env_var;
use anyhow::Context;
use aws_sdk_s3 as s3;
use std::fs::File;
use std::io::prelude::{Read, Write};
use std::path::{Path, PathBuf};

pub enum Storage {
    LocalFiles { dir: PathBuf },
    S3 { client: s3::Client, bucket: String },
}

use Storage::*;

impl Storage {
    pub fn from_local_dir(dir: PathBuf) -> Self {
        LocalFiles { dir }
    }

    pub async fn from_s3(bucket: String) -> Self {
        let sdk_config = aws_config::load_from_env().await;
        let config = aws_sdk_s3::config::Builder::from(&sdk_config)
            .force_path_style(true)
            .build();
        let client = aws_sdk_s3::Client::from_conf(config);
        S3 { client, bucket }
    }

    pub async fn from_env() -> anyhow::Result<Self> {
        match env_var("STORAGE_BACKEND")?.as_ref() {
            "local" => {
                let dir = match env_var("STORAGE_LOCAL_DIR") {
                    Ok(dir) => dir,
                    Err(_) => ".".to_string(),
                };
                Ok(Self::from_local_dir(PathBuf::from(dir)))
            }
            "s3" => {
                let bucket = env_var("STORAGE_S3_BUCKET")?;
                Ok(Self::from_s3(bucket).await)
            }
            other => {
                anyhow::bail!("STORAGE_BACKEND: unexpected value {other} (expected: local/s3)")
            }
        }
    }

    fn read_file(dir: &Path, path: &str) -> anyhow::Result<Option<Vec<u8>>> {
        let path = dir.join(path);
        if path.is_file() {
            let mut file = File::open(path)?;
            let mut data = vec![];
            file.read_to_end(&mut data)?;
            Ok(Some(data))
        } else {
            Ok(None)
        }
    }

    fn write_file(dir: &Path, path: &str, data: Vec<u8>) -> anyhow::Result<()> {
        let path = dir.join(path);
        let mut file = File::create(path)?;
        file.write_all(&data)?;
        Ok(())
    }

    async fn read_s3(
        client: &s3::Client,
        bucket: &str,
        path: &str,
    ) -> anyhow::Result<Option<Vec<u8>>> {
        let output = client.get_object().bucket(bucket).key(path).send().await;
        match output {
            Ok(output) => {
                let data = output
                    .body
                    .collect()
                    .await
                    .context(format!("error reading {} from bucket {}", path, bucket))?
                    .into_bytes()
                    .to_vec();
                Ok(Some(data))
            }
            Err(e) if is_no_such_key_error(&e) => Ok(None),
            Err(err) => Err(err)?,
        }
    }

    async fn write_s3(
        client: &s3::Client,
        bucket: &str,
        path: &str,
        data: Vec<u8>,
    ) -> anyhow::Result<()> {
        client
            .put_object()
            .bucket(bucket)
            .key(path)
            .body(s3::primitives::ByteStream::from(data))
            .send()
            .await?;
        Ok(())
    }

    pub async fn write(&self, path: &str, data: Vec<u8>) -> anyhow::Result<()> {
        match self {
            LocalFiles { dir } => Self::write_file(dir, path, data),
            S3 { client, bucket } => Self::write_s3(client, bucket, path, data).await,
        }
    }

    pub async fn read(&self, path: &str) -> anyhow::Result<Option<Vec<u8>>> {
        match self {
            LocalFiles { dir } => Self::read_file(dir, path),
            S3 { client, bucket } => Self::read_s3(client, bucket, path).await,
        }
    }
}

use s3::error::SdkError;
use s3::operation::get_object::GetObjectError;

fn is_no_such_key_error<R>(err: &SdkError<GetObjectError, R>) -> bool {
    match err {
        SdkError::ServiceError(e) => matches!(e.err(), GetObjectError::NoSuchKey(_)),
        _ => false,
    }
}

use crate::classifier::Classifier;

pub async fn load_classifier(storage: &Storage) -> anyhow::Result<Classifier> {
    match storage.read("model.json").await? {
        Some(data) => Ok(serde_json::from_slice(&data)?),
        None => Ok(Classifier::new()),
    }
}

use crate::userdb::UserDb;

pub async fn load_userdb(storage: &Storage) -> anyhow::Result<Option<UserDb>> {
    if let Some(data) = storage.read("db.json").await? {
        Ok(Some(serde_json::from_slice(&data)?))
    } else {
        Ok(None)
    }
}

pub async fn store_userdb(storage: &Storage, userdb: &UserDb) -> anyhow::Result<()> {
    storage
        .write("db.json", serde_json::to_vec(userdb)?)
        .await?;
    Ok(())
}

use crate::db::Db;

pub async fn store_db(storage: &Storage, db: &Db) -> anyhow::Result<()> {
    let userdb_bytes = db.with_userdb(serde_json::to_vec)?;
    let classifier_bytes = db.with_classifier(serde_json::to_vec)?;
    storage.write("db.json", userdb_bytes).await?;
    storage.write("model.json", classifier_bytes).await?;
    Ok(())
}
